"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class NumericDatapoint(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TIMESTAMP_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    timestamp: builtins.int
    value: builtins.float
    def __init__(
        self,
        *,
        timestamp: builtins.int = ...,
        value: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["timestamp", b"timestamp", "value", b"value"]) -> None: ...

global___NumericDatapoint = NumericDatapoint

@typing_extensions.final
class NumericDatapoints(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATAPOINTS_FIELD_NUMBER: builtins.int
    @property
    def datapoints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___NumericDatapoint]: ...
    def __init__(
        self,
        *,
        datapoints: collections.abc.Iterable[global___NumericDatapoint] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["datapoints", b"datapoints"]) -> None: ...

global___NumericDatapoints = NumericDatapoints

@typing_extensions.final
class StringDatapoint(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TIMESTAMP_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    timestamp: builtins.int
    value: builtins.str
    def __init__(
        self,
        *,
        timestamp: builtins.int = ...,
        value: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["timestamp", b"timestamp", "value", b"value"]) -> None: ...

global___StringDatapoint = StringDatapoint

@typing_extensions.final
class StringDatapoints(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATAPOINTS_FIELD_NUMBER: builtins.int
    @property
    def datapoints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___StringDatapoint]: ...
    def __init__(
        self,
        *,
        datapoints: collections.abc.Iterable[global___StringDatapoint] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["datapoints", b"datapoints"]) -> None: ...

global___StringDatapoints = StringDatapoints

@typing_extensions.final
class AggregateDatapoint(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TIMESTAMP_FIELD_NUMBER: builtins.int
    AVERAGE_FIELD_NUMBER: builtins.int
    MAX_FIELD_NUMBER: builtins.int
    MIN_FIELD_NUMBER: builtins.int
    COUNT_FIELD_NUMBER: builtins.int
    SUM_FIELD_NUMBER: builtins.int
    INTERPOLATION_FIELD_NUMBER: builtins.int
    STEPINTERPOLATION_FIELD_NUMBER: builtins.int
    CONTINUOUSVARIANCE_FIELD_NUMBER: builtins.int
    DISCRETEVARIANCE_FIELD_NUMBER: builtins.int
    TOTALVARIATION_FIELD_NUMBER: builtins.int
    timestamp: builtins.int
    average: builtins.float
    max: builtins.float
    min: builtins.float
    count: builtins.float
    sum: builtins.float
    interpolation: builtins.float
    stepInterpolation: builtins.float
    continuousVariance: builtins.float
    discreteVariance: builtins.float
    totalVariation: builtins.float
    def __init__(
        self,
        *,
        timestamp: builtins.int = ...,
        average: builtins.float = ...,
        max: builtins.float = ...,
        min: builtins.float = ...,
        count: builtins.float = ...,
        sum: builtins.float = ...,
        interpolation: builtins.float = ...,
        stepInterpolation: builtins.float = ...,
        continuousVariance: builtins.float = ...,
        discreteVariance: builtins.float = ...,
        totalVariation: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["average", b"average", "continuousVariance", b"continuousVariance", "count", b"count", "discreteVariance", b"discreteVariance", "interpolation", b"interpolation", "max", b"max", "min", b"min", "stepInterpolation", b"stepInterpolation", "sum", b"sum", "timestamp", b"timestamp", "totalVariation", b"totalVariation"]) -> None: ...

global___AggregateDatapoint = AggregateDatapoint

@typing_extensions.final
class AggregateDatapoints(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DATAPOINTS_FIELD_NUMBER: builtins.int
    @property
    def datapoints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___AggregateDatapoint]: ...
    def __init__(
        self,
        *,
        datapoints: collections.abc.Iterable[global___AggregateDatapoint] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["datapoints", b"datapoints"]) -> None: ...

global___AggregateDatapoints = AggregateDatapoints
