from __future__ import annotations

LIST_LIMIT_DEFAULT = 25
# Max JavaScript-safe integer 2^53 - 1
MAX_VALID_INTERNAL_ID = 9007199254740991
DATA_MODELING_LIST_LIMIT_DEFAULT = 10
INSTANCES_LIST_LIMIT_DEFAULT = 1_000
DATAPOINT_SUBSCRIPTIONS_LIST_LIMIT_DEFAULT = 100
DATAPOINT_SUBSCRIPTION_DATA_LIST_LIMIT_DEFAULT = 100
